# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import numpy as np
from typing import Union, Iterable
from ..core import Plot, Array
from .render import render
from ..core.tools import to_bin_centers, finmin, finmax


def histogram1d(
    *layers: Union[Iterable, Array],
    bins: Union[int, Iterable] = 50,
    weights: Array = None,
    logx: bool = False,
    logy: bool = False,
    loglog: bool = False,
    filename: str = None,
    title: str = None,
    ymin: float = None,
    ymax: float = None,
    ax: object = None,
    **kwargs,
) -> Plot:
    """
    Plot a 1D histogram with arbitrary number of variables as input.
    When a vector quantity is supplied, the function will histogram the norm of
    the vectors.


    This function has an API very close to that of matplotlib's ``hist`` function.
    For the documentation of any parameters that are not listed below, see
    https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.hist.html.

    :param layers: Dicts or Arrays representing the quantities to be mapped onto the
        colormap of the generated image.

    :param logx: If ``True``, use logarithmic scaling on the horizontal axis.
        Default is ``False``.

    :param logy: If ``True``, use logarithmic scaling on the vertical axis.
        Default is ``False``.

    :param loglog: If ``True``, use logarithmic scaling on the horizontal and
        vertical axes. Default is ``False``.

    :param filename: If specified, the returned figure is also saved to file.
        Default is ``None``.

    :param title: The title of the figure. Default is ``None``.

    :param ymin: Minimum value for the vertical axis. Default is ``None``.

    :param ymax: Maximum value for the vertical axis. Default is ``None``.

    :param ax: A matplotlib axes inside which the figure will be plotted.
        Default is ``None``, in which case some new axes a created.
    """
    if loglog:
        logx = logy = True

    figure = render(logx=logx, logy=logy, ax=ax)

    for layer in layers:
        if isinstance(layer, dict):
            params = {}
            extra_args = {}
            for key, param in layer.items():
                if key in ["data", "bins", "weights"]:
                    params[key] = param
                else:
                    extra_args[key] = param
            for key, arg in {"bins": bins, "weights": weights}.items():
                if key not in params:
                    params[key] = arg
        else:
            params = {"data": layer, "bins": bins, "weights": weights}
            extra_args = kwargs

        xvals = params["data"].norm.values
        if params["weights"] is not None:
            params["weights"] = params["weights"].norm.values

        # Construct some bin edges
        if isinstance(params["bins"], int):
            xmin = finmin(xvals)
            xmax = finmax(xvals)
            if logx:
                xedges = np.logspace(np.log10(xmin), np.log10(xmax), params["bins"] + 1)
            else:
                xedges = np.linspace(xmin, xmax, params["bins"] + 1)
        else:
            xedges = params["bins"]

        ydata, _, _ = figure["ax"].hist(
            xvals, bins=xedges, weights=params["weights"], **extra_args
        )

        figure["ax"].set_xlabel(params["data"].label)

    figure["ax"].set_ylim(ymin, ymax)
    return Plot(
        x=to_bin_centers(xedges),
        y=ydata,
        fig=figure["fig"],
        ax=figure["ax"],
        filename=filename,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
import numpy as np
from copy import copy, deepcopy
from .. import config
from ..io import Loader
from .datagroup import Datagroup
from .tools import bytes_to_human_readable
from ..units import units, UnitsLibrary


class Dataset:
    def __init__(self, nout=None, path=""):
        self.groups = {}
        self.meta = {}
        self.loader = None
        self.units = None
        if nout is not None:
            self.loader = Loader(nout=nout, path=path)
            self.meta.update(self.loader.load_metadata())
            self.set_units()
            self.meta["time"] *= self.units["time"]

    def __iter__(self):
        return self.groups.__iter__()

    def __len__(self):
        return self.groups.__len__()

    def __getitem__(self, key):
        return self.groups[key]

    def __setitem__(self, key, value):
        if not isinstance(value, Datagroup):
            raise TypeError(
                "Only objects of type Datagroup can be inserted into a " "Dataset."
            )
        self.groups.__setitem__(key, value)
        value.name = key
        value.parent = self

    def __delitem__(self, key):
        return self.groups.__delitem__(key)

    def __repr__(self):
        return str(self)

    def __str__(self):
        output = "Dataset: "
        if "infile" in self.meta:
            output += "{}: ".format(self.meta["infile"])
        output += "{}\n".format(self.print_size())
        for key, item in self.items():
            output += str(item) + "\n"
        return output

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        nout = self.loader.nout if self.loader is not None else None
        path = self.loader.path if self.loader is not None else ""
        out = self.__class__(nout=nout, path=path)
        for key, group in self.groups.items():
            out[key] = deepcopy(group)
        out.meta = {key: copy(item) for key, item in self.meta.items()}
        return out

    def copy(self):
        nout = self.loader.nout if self.loader is not None else None
        path = self.loader.path if self.loader is not None else ""
        out = self.__class__(nout=nout, path=path)
        for key, group in self.groups.items():
            out[key] = group
        out.meta = self.meta.copy()
        return out

    def keys(self):
        return self.groups.keys()

    def items(self):
        return self.groups.items()

    def values(self):
        return self.groups.values()

    def load(self, *args, **kwargs):
        groups = self.loader.load(*args, meta=self.meta, units=self.units, **kwargs)
        for name, group in groups.items():
            self[name] = group
        config.additional_variables(self)
        return self

    def nbytes(self):
        return np.sum([item.nbytes() for item in self.groups.values()])

    def print_size(self):
        return bytes_to_human_readable(self.nbytes())

    def clear(self):
        self.groups.clear()
        self.meta.clear()

    def get(self, key, default):
        return self.groups.get(key, default)

    def pop(self, key):
        return self.groups.pop(key)

    def update(self, d):
        for key, value in d.items():
            self[key] = value

    def set_units(self):
        self.units = UnitsLibrary(
            library=config.configure_units(
                units=units,
                unit_d=self.meta["unit_d"],
                unit_l=self.meta["unit_l"],
                unit_t=self.meta["unit_t"],
            ),
            default_unit=1.0 * units(""),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import numpy as np
from numba import njit, prange


@njit(parallel=True)
def evaluate_on_grid(
    cell_positions_in_new_basis_x,
    cell_positions_in_new_basis_y,
    cell_positions_in_new_basis_z,
    cell_positions_in_original_basis_x,
    cell_positions_in_original_basis_y,
    cell_positions_in_original_basis_z,
    cell_values,
    cell_sizes,
    grid_lower_edge_in_new_basis_x,
    grid_lower_edge_in_new_basis_y,
    grid_lower_edge_in_new_basis_z,
    grid_spacing_in_new_basis_x,
    grid_spacing_in_new_basis_y,
    grid_spacing_in_new_basis_z,
    grid_positions_in_original_basis,
    ndim,
):
    nz, ny, nx = grid_positions_in_original_basis.shape[:3]
    diagonal = np.sqrt(ndim)
    out = np.full(
        shape=(cell_values.shape[0], nz, ny, nx), fill_value=np.nan, dtype=np.float64
    )

    ncells = len(cell_positions_in_new_basis_x)
    for n in prange(ncells):
        half_size = cell_sizes[n] * diagonal
        ix1 = max(
            int(
                (
                    (cell_positions_in_new_basis_x[n] - half_size)
                    - grid_lower_edge_in_new_basis_x
                )
                / grid_spacing_in_new_basis_x
            ),
            0,
        )
        ix2 = min(
            int(
                (
                    (cell_positions_in_new_basis_x[n] + half_size)
                    - grid_lower_edge_in_new_basis_x
                )
                / grid_spacing_in_new_basis_x
            )
            + 1,
            nx,
        )
        iy1 = max(
            int(
                (
                    (cell_positions_in_new_basis_y[n] - half_size)
                    - grid_lower_edge_in_new_basis_y
                )
                / grid_spacing_in_new_basis_y
            ),
            0,
        )
        iy2 = min(
            int(
                (
                    (cell_positions_in_new_basis_y[n] + half_size)
                    - grid_lower_edge_in_new_basis_y
                )
                / grid_spacing_in_new_basis_y
            )
            + 1,
            ny,
        )
        iz1 = max(
            int(
                (
                    (cell_positions_in_new_basis_z[n] - half_size)
                    - grid_lower_edge_in_new_basis_z
                )
                / grid_spacing_in_new_basis_z
            ),
            0,
        )
        iz2 = min(
            int(
                (
                    (cell_positions_in_new_basis_z[n] + half_size)
                    - grid_lower_edge_in_new_basis_z
                )
                / grid_spacing_in_new_basis_z
            )
            + 1,
            nz,
        )

        for k in range(iz1, iz2):
            for j in range(iy1, iy2):
                for i in range(ix1, ix2):
                    ok_x = (
                        np.abs(
                            grid_positions_in_original_basis[k, j, i, 0]
                            - cell_positions_in_original_basis_x[n]
                        )
                        <= cell_sizes[n]
                    )
                    ok_y = True
                    if cell_positions_in_original_basis_y is not None:
                        ok_y = (
                            np.abs(
                                grid_positions_in_original_basis[k, j, i, 1]
                                - cell_positions_in_original_basis_y[n]
                            )
                            <= cell_sizes[n]
                        )
                    ok_z = True
                    if cell_positions_in_original_basis_z is not None:
                        ok_z = (
                            np.abs(
                                grid_positions_in_original_basis[k, j, i, 2]
                                - cell_positions_in_original_basis_z[n]
                            )
                            <= cell_sizes[n]
                        )

                    if ok_x and ok_y and ok_z:
                        out[:, k, j, i] = cell_values[:, n]

    return out


@njit(parallel=True)
def hist2d(x, y, values, xmin, xmax, nx, ymin, ymax, ny):
    out = np.zeros(shape=(values.shape[0], ny, nx), dtype=np.float64)
    counts = np.zeros(shape=(ny, nx), dtype=np.int64)
    dx = (xmax - xmin) / nx
    dy = (ymax - ymin) / ny

    for i in prange(len(x)):
        indx = int((x[i] - xmin) / dx)
        indy = int((y[i] - ymin) / dy)
        if (indx >= 0) and (indx < nx) and (indy >= 0) and (indy < ny):
            out[:, indy, indx] += values[:, i]
            counts[indy, indx] += 1

    return out, counts



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
import numpy as np
from pint import Quantity
from pint.errors import DimensionalityError
from .base import Base
from .tools import value_to_string
from .. import units

APPLY_OP_TO_UNIT = ("multiply", "true_divide", "divide", "sqrt", "power", "reciprocal")


def _binary_op(op, lhs, rhs, strict=True, **kwargs):
    if not isinstance(rhs, lhs.__class__):
        try:
            rhs = lhs.__class__(rhs)
        except NotImplementedError:
            return NotImplemented
    if strict:
        rhs = rhs.to(lhs.unit)
    else:
        try:
            rhs = rhs.to(lhs.unit)
        except DimensionalityError:
            pass
    return op(lhs, rhs, **kwargs)


class Array(Base):
    def __init__(self, values, unit=None, parent=None, name=""):
        if isinstance(values, Base):
            raise NotImplementedError("Cannot create Array from Array or Vector.")

        if isinstance(values, Quantity):
            if unit is not None:
                raise ValueError(
                    "Cannot set unit when creating an Array from a Quantity."
                )
            self._array = values.magnitude
            self.unit = values.units
        else:
            self._array = values
            self.unit = units(unit)
        if not isinstance(self._array, np.ndarray):
            self._array = np.asarray(self._array)

        self.parent = parent
        self.name = name

    def __getitem__(self, slice_):
        if isinstance(slice_, Base):
            if not isinstance(slice_, self.__class__):
                raise ValueError(
                    "Cannot slice using a Vector, only Array is supported."
                )
            if slice_.dtype not in ("int32", "int64", bool):
                raise TypeError(
                    "Dtype of the Array must be integer or bool for slicing."
                )
            slice_ = slice_.values
        return self.__class__(
            values=self._array[slice_],
            unit=self.unit,
            parent=self.parent,
            name=self.name,
        )

    def __len__(self):
        if self._array.shape:
            return len(self._array)
        else:
            return 0

    def __str__(self):
        name_str = "'" + self.name + "' "
        if len(self) == 0:
            values_str = "Value: " + value_to_string(self.values)
        else:
            values_str = (
                "Min: "
                + value_to_string(self.min().values)
                + " Max: "
                + value_to_string(self.max().values)
            )
        unit_str = " [{:~}] ".format(self.unit)
        shape_str = str(self.shape)
        return name_str + values_str + unit_str + shape_str

    def copy(self):
        return self.__class__(
            values=self._array.copy(), unit=units(self.unit), name=str(self.name)
        )

    @property
    def values(self):
        if not self._array.shape:
            return self._array[()]
        else:
            return self._array

    @values.setter
    def values(self, values_):
        self._array = values_

    @property
    def norm(self):
        return self

    @property
    def ndim(self):
        return self._array.ndim

    @property
    def shape(self):
        return self._array.shape

    @property
    def dtype(self):
        return self._array.dtype

    def __add__(self, other):
        return _binary_op(np.add, self, other)

    def __iadd__(self, other):
        return _binary_op(np.add, self, other, out=self)

    def __sub__(self, other):
        return _binary_op(np.subtract, self, other)

    def __isub__(self, other):
        return _binary_op(np.subtract, self, other, out=self)

    def __mul__(self, other):
        return _binary_op(np.multiply, self, other, strict=False)

    def __imul__(self, other):
        return _binary_op(np.multiply, self, other, strict=False, out=self)

    def __truediv__(self, other):
        return _binary_op(np.divide, self, other, strict=False)

    def __itruediv__(self, other):
        return _binary_op(np.divide, self, other, strict=False, out=self)

    def __rmul__(self, other):
        return self * other

    def __rtruediv__(self, other):
        return np.reciprocal(self / other)

    def __pow__(self, number):
        return np.power(self, number)

    def __neg__(self):
        return np.negative(self)

    def __lt__(self, other):
        return _binary_op(np.less, self, other)

    def __le__(self, other):
        return _binary_op(np.less_equal, self, other)

    def __gt__(self, other):
        return _binary_op(np.greater, self, other)

    def __ge__(self, other):
        return _binary_op(np.greater_equal, self, other)

    def __eq__(self, other):
        return _binary_op(np.equal, self, other)

    def __ne__(self, other):
        return _binary_op(np.not_equal, self, other)

    def __and__(self, other):
        return _binary_op(np.logical_and, self, other)

    def __or__(self, other):
        return _binary_op(np.logical_or, self, other)

    def __xor__(self, other):
        return _binary_op(np.logical_xor, self, other)

    def __invert__(self):
        return np.logical_not(self)

    def to(self, unit):
        new_unit = units(unit)
        if self.unit == new_unit:
            return self
        ratio = (1.0 * self.unit).to(new_unit) / (1.0 * new_unit)
        return self.__class__(values=self._array * ratio.magnitude, unit=new_unit)

    def _maybe_array(self, arg):
        if isinstance(arg, self.__class__):
            return arg._array
        if isinstance(arg, Quantity):
            return arg.magnitude
        return arg

    def _extract_arrays_from_args(self, args):
        return tuple(self._maybe_array(a) for a in args)

    def _extract_arrays_from_kwargs(self, kwargs):
        return {key: self._extract_arrays_from_args(a) for key, a in kwargs.items()}

    def _maybe_unit(self, arg):
        if hasattr(arg, "unit"):
            return 1.0 * arg.unit
        if hasattr(arg, "units"):
            return 1.0 * arg.units
        return arg

    def _extract_units(self, args):
        return tuple(self._maybe_unit(a) for a in args)

    def _wrap_numpy(self, func, *args, **kwargs):
        if isinstance(args[0], (tuple, list)):
            array_args = (
                self._extract_arrays_from_args(args[0]),
            ) + self._extract_arrays_from_args(args[1:])
        else:
            array_args = self._extract_arrays_from_args(args)
        result = func(*array_args, **self._extract_arrays_from_kwargs(kwargs))

        unit = None
        if result.dtype in (int, float):
            if func.__name__ in APPLY_OP_TO_UNIT:
                unit = func(
                    *self._extract_units(args),
                    **{key: a for key, a in kwargs.items() if key != "out"},
                ).units
            else:
                unit = self.unit

        if "out" in kwargs:
            kwargs["out"][0].unit = unit
            return kwargs["out"][0]
        else:
            return self.__class__(values=result, unit=unit)

    def reshape(self, *shape):
        return self.__class__(values=self._array.reshape(*shape), unit=self.unit)

    @property
    def nbytes(self):
        return self._array.nbytes



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
import numpy as np
from .tools import bytes_to_human_readable


class Datagroup:
    def __init__(self, data=None, parent=None):
        self._container = {}
        self.parent = parent
        self.name = ""
        if data is not None:
            for key, array in data.items():
                self[key] = array

    def __iter__(self):
        return self._container.__iter__()

    def __len__(self):
        return self._container.__len__()

    def __getitem__(self, key):
        if isinstance(key, str):
            return self._container[key]
        else:
            if isinstance(key, int):
                key = slice(key, key + 1, 1)
            d = self.__class__()
            for name, val in self.items():
                d[name] = val[key]
            return d

    def __setitem__(self, key, value):
        if self.shape and (self.shape != value.shape):
            raise ValueError(
                "Size mismatch on element insertion. Item "
                "shape is {} while container accepts shape {}.".format(
                    value.shape, self.shape
                )
            )
        value.name = key
        value.parent = self
        self._container[key] = value

    def __delitem__(self, key):
        return self._container.__delitem__(key)

    def __repr__(self):
        return str(self)

    def __str__(self):
        output = "Datagroup: {} {}\n".format(self.name, self.print_size())
        for key, item in self.items():
            output += str(item) + "\n"
        return output

    def __eq__(self, other):
        if self.keys() != other.keys():
            return False
        for key, value in self.items():
            if all(value != other[key]):
                return False
        return True

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.__class__(data={key: array.copy() for key, array in self.items()})

    def copy(self):
        return self.__class__(data=self._container.copy())

    def keys(self):
        return self._container.keys()

    def items(self):
        return self._container.items()

    def values(self):
        return self._container.values()

    def nbytes(self):
        return np.sum([item.nbytes for item in self.values()])

    def print_size(self):
        return bytes_to_human_readable(self.nbytes())

    @property
    def shape(self):
        if len(self) == 0:
            return ()
        else:
            return self[list(self.keys())[0]].shape

    def sortby(self, key):
        if key is not None:
            if isinstance(key, str):
                key = np.argsort(self[key]).values
            for var in self.keys():
                self[var] = self[var][key]

    def clear(self):
        self._container.clear()

    def get(self, key, default):
        return self._container.get(key, default)

    def pop(self, key):
        return self._container.pop(key)

    def update(self, d):
        for key, value in d.items():
            self[key] = value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import numpy as np


def arrayclose(a, b):
    if a.unit != b.unit:
        return False
    return np.allclose(a.values, b.values)


def arraytrue(a):
    return all(a.values)


def arrayequal(a, b):
    if a.unit != b.unit:
        return False
    return np.array_equal(a.values, b.values)


def vectorclose(a, b):
    ok = arrayclose(a.x, b.x)
    if a.y is not None:
        ok *= arrayclose(a.y, b.y)
    if a.z is not None:
        ok *= arrayclose(a.z, b.z)
    return bool(ok)


def vectortrue(a):
    ok = arraytrue(a.x)
    if a.y is not None:
        ok *= arraytrue(a.y)
    if a.z is not None:
        ok *= arraytrue(a.z)
    return bool(ok)


def vectorequal(a, b):
    ok = arrayequal(a.x, b.x)
    if a.y is not None:
        ok *= arrayequal(a.y, b.y)
    if a.z is not None:
        ok *= arrayequal(a.z, b.z)
    return bool(ok)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import numpy as np
from .. import Dataset
import warnings


def extract_sphere(dataset, radius, origin):
    """
    Extract a spherical subdomain around an origin point.
    """
    subdomain = Dataset()
    subdomain.meta = dataset.meta.copy()

    for name, group in dataset.items():
        pos = group.get("position", group.parent["amr"]["position"])
        if pos.shape != group.shape:
            warnings.warn(
                "Ignoring datagroup '{}', which has no position ".format(group)
                + "vector and has different shape than 'amr' group."
            )
            continue
        r = (pos - origin).norm
        c = (r < radius).values
        if np.any(c):
            subdomain[name] = dataset[name][c]

    return subdomain


def extract_box(dataset, dx, dy, dz, origin):
    """
    Extract a cubic domain of size dx, dy & dz around an origin point
    """
    subdomain = Dataset()
    subdomain.meta = dataset.meta.copy()

    for name, group in dataset.items():
        pos = group.get("position", group.parent["amr"]["position"])
        if pos.shape != group.shape:
            warnings.warn(
                "Ignoring datagroup '{}', which has no position ".format(group)
                + "vector and has different shape than 'amr' group."
            )
            continue
        centered_pos = pos - origin
        cx = (centered_pos.x <= dx * 0.5) & (centered_pos.x >= -dx * 0.5)
        cy = (centered_pos.y <= dy * 0.5) & (centered_pos.y >= -dy * 0.5)
        cz = (centered_pos.z <= dz * 0.5) & (centered_pos.z >= -dz * 0.5)
        c = (cx & cy & cz).values
        if np.any(c):
            subdomain[name] = dataset[name][c]

    return subdomain



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

# flake8: noqa

from .config import config
from .units import units
from .core import Array, Datagroup, Dataset, Plot, Vector
from .plot import histogram1d, histogram2d, scatter, map, plot
from .spatial import extract_box, extract_sphere



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
"""
Define default values so that you don't have to specify them every time.
"""

from math import sqrt, pi


def configure_constants(units):
    units.define("bolometric_luminosity = 3.0128e+28 * W = L_bol0")
    units.define("solar_luminosity = 3.828e+26 * W = L_sun = L_sol")
    units.define("earth_mass = 5.97216787e+27 * g = M_earth")
    units.define("jupiter_mass = 1.8981246e+30 * g = M_jup")
    units.define("solar_mass = 1.9889e+33 * g = M_sun = M_sol")
    units.define("earth_radius = 6.3781e+08 * cm = R_earth")
    units.define("jupiter_radius = 7.1492e+09 * cm = R_jup")
    units.define("solar_radius = 6.957e+10 * cm = R_sun = R_sol")
    units.define(
        "radiation_constant = 7.56591469318689378e-015 * erg / cm^3 / K^4 = ar"
    )


def configure_units(units, unit_d, unit_l, unit_t):
    density = unit_d * units("g / cm**3")
    velocity = (unit_l / unit_t) * units("cm / s")
    magnetic_field = sqrt(4.0 * pi * unit_d * (unit_l / unit_t) ** 2) * units("G")
    momentum = density * velocity
    acceleration = (unit_l / unit_t**2) * units("cm / s**2")
    energy = unit_d * ((unit_l / unit_t) ** 2) * units("erg / cm**3")
    time = unit_t * units("s")
    length = unit_l * units("cm")
    mass = density * length**3
    temperature = 1.0 * units("K")
    grav_potential = velocity**2

    library = {
        "unit_d": unit_d,
        "unit_l": unit_l,
        "unit_t": unit_t,
        "density": density,
        "velocity": velocity,
        "velocity_*": velocity,
        "momentum": momentum,
        "momentum_*": momentum,
        "magnetic_field": magnetic_field,
        "B_left": magnetic_field,
        "B_left_*": magnetic_field,
        "B_right": magnetic_field,
        "B_right_*": magnetic_field,
        "B_field": magnetic_field,
        "B_field_*": magnetic_field,
        "B_*_left": magnetic_field,
        "B_*_right": magnetic_field,
        "acceleration": acceleration,
        "grav_acceleration": acceleration,
        "grav_acceleration_*": acceleration,
        "grav_potential": grav_potential,
        "energy": energy,
        "internal_energy": energy,
        "thermal_pressure": energy,
        "pressure": energy,
        "radiative_energy": energy,
        "radiative_energy_*": energy,
        "time": time,
        "length": length,
        "x": length,
        "y": length,
        "z": length,
        "position": length,
        "position_*": length,
        "dx": length,
        "mass": mass,
        "temperature": temperature,
    }
    return library


def additional_variables(data):
    """
    Here are some additional variables that are to be computed every time data
    is loaded.

    It is recommended to place your variables in a `try/except` block, which
    will prevent errors if the variables are not found, for instance when
    loading data from a different simulation.
    """

    # Magnetic field
    try:
        data["hydro"]["B_field"] = 0.5 * (
            data["hydro"]["B_left"] + data["hydro"]["B_right"]
        )
    except KeyError:
        pass

    # Mass
    try:
        data["hydro"]["mass"] = (data["hydro"]["density"] * data["amr"]["dx"] ** 3).to(
            "M_sun"
        )
    except KeyError:
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

import re


class UnitsLibrary:
    """
    A small helper class that holds the list of units, and behaves a bit like a
    dict but also performs regex matching on key with wildcard character '*'.
    """

    def __init__(self, library, default_unit):
        self._library = library
        self._default_unit = default_unit

    def __getitem__(self, key):
        if key in self._library:
            return self._library[key]
        for name, unit in self._library.items():
            if "*" in name:
                regex = re.compile(name.replace("*", ".+"))
                if re.match(regex, key):
                    return unit
        return self._default_unit

    def __setitem__(self, key, value):
        self._library[key] = value

    def __str__(self):
        return str(self._library)

    def __repr__(self):
        return str(self)

    def keys(self):
        return self._library.keys()

    def values(self):
        return self._library.values()

    def items(self):
        return self._library.items()

    def update(self, *args, **kwargs):
        self._library.update(*args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)

from ..core import Vector

import numpy as np


def _perpendicular_vector(v):
    """
    Compute a vector perpendicular to the input vector
    """

    if v.z.values == 0:
        return Vector(-v.y.values, v.x.values, 0, unit=v.unit)
    else:
        return Vector(1.0, 1.0, (-1.0 * (v.x + v.y) / v.z).values, unit=v.unit)


def get_direction(direction=None, dataset=None, dx=None, dy=None, origin=None):
    """
    Find direction vectors for slice.

    Direction can be:

    - a single letter ``"x"``, ``"y"``, ``"z"`` representing the normal to the plane
    - three letters, e.g. ``"zyx"``, where the first letter is the normal to the plane,
      and the following two letters are the two other perpendicual vectors
    - a list of 3 numbers, representing the 3 components of the normal to the plane
    - ``"top"`` for a 'top' view according to local angular momentum
    - ``"side"`` for a 'side' view according to local angular momentum
    - a list of 2 lists of 3 numbers, representing the normal to the plane and the first
      of the two other perpendicular vectors (the third one will be computed from the
      other two)

    The origin is a vector of 3 numbers (xyz).
    """
    dir_list = {
        "x": Vector(1, 0, 0, name="x"),
        "y": Vector(0, 1, 0, name="y"),
        "z": Vector(0, 0, 1, name="z"),
    }
    dir_names = ("normal", "pos_u", "pos_v")

    if dataset.meta["ndim"] < 3:
        return {
            dir_names[0]: Vector(0, 0, name="z"),
            dir_names[1]: Vector(1, 0, name="x"),
            dir_names[2]: Vector(0, 1, name="y"),
        }

    if isinstance(direction, str):
        direction = direction.lower()
        if direction in ["top", "side"]:
            if dx is None:
                pos = dataset["amr"]["position"]
                sphere_rad = (
                    0.5
                    * (
                        pos.x.max()
                        - pos.x.min()
                        + pos.y.max()
                        - pos.y.min()
                        + pos.z.max()
                        - pos.z.min()
                    )
                    / 3.0
                ) * dataset["amr"]["position"].unit
            else:
                sphere_rad = 0.25 * (dx + dy)

            xyz = dataset["amr"]["position"]
            if origin is not None:
                xyz = xyz - origin
            # Compute angular momentum vector
            sphere = (xyz.norm < sphere_rad).values
            pos = xyz * dataset["hydro"]["mass"]
            vel = dataset["hydro"]["velocity"]

            ang_mom = np.sum(pos[sphere].cross(vel[sphere]))
            if direction == "side":
                # Choose a vector perpendicular to the angular momentum vector
                dir3 = ang_mom
                dir1 = _perpendicular_vector(dir3)
                dir2 = dir3.cross(dir1)
            else:
                dir1 = ang_mom
                dir2 = _perpendicular_vector(dir1)
                dir3 = dir1.cross(dir2)
            dir_vecs = {}
            print("Basis vectors:")
            for key, vec in zip(dir_names, (dir1, dir2, dir3)):
                vec.name = key
                dir_vecs[key] = vec
                print(vec)

        if set(direction) == set("xyz"):  # case where direction = "xyz", "zyx" etc.
            return {
                "normal": dir_list[direction[0]],
                "pos_u": dir_list[direction[1]],
                "pos_v": dir_list[direction[2]],
            }
        elif direction == "x":
            return {
                "normal": dir_list["x"],
                "pos_u": dir_list["y"],
                "pos_v": dir_list["z"],
            }
        elif direction == "y":
            return {
                "normal": dir_list["y"],
                "pos_u": dir_list["z"],
                "pos_v": dir_list["x"],
            }
        elif direction == "z":
            return {
                "normal": dir_list["z"],
                "pos_u": dir_list["x"],
                "pos_v": dir_list["y"],
            }
    elif isinstance(direction, Vector):
        dir_vecs = {"normal": direction}
        u = _perpendicular_vector(dir_vecs["normal"])
        u.name = "pos_u"
        dir_vecs["pos_u"] = u
        v = direction.cross(dir_vecs["pos_u"])
        v.name = "pos_v"
        dir_vecs["pos_v"] = v
    elif isinstance(direction, dict):
        dir_vecs = direction.copy()
        assert "normal" in dir_vecs
        assert "pos_u" in dir_vecs
        assert "pos_v" in dir_vecs
    else:
        raise ValueError(f"Bad direction for slice: {direction}.")

    for key, vec in dir_vecs.items():
        v = vec / vec.norm
        v.name = vec.name
        dir_vecs[key] = v

    return dir_vecs



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
from common import arrayequal
from osyris import Array, Datagroup
from copy import copy, deepcopy
import pytest


def test_datagroup_creation():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup()
    dg["a"] = a
    dg["b"] = b
    assert "a" in dg.keys()
    assert "b" in dg.keys()
    assert dg["a"].name == "a"
    assert dg["b"].name == "b"
    assert len(dg) == 2


def test_datagroup_creation_from_dict():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    assert "a" in dg.keys()
    assert "b" in dg.keys()
    assert dg["a"].name == "a"
    assert dg["b"].name == "b"
    assert len(dg) == 2


def test_datagroup_bad_length_insertion():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0], unit="s")
    dg = Datagroup()
    dg["a"] = a
    with pytest.raises(ValueError):
        dg["b"] = b


def test_datagroup_delitem():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    del dg["a"]
    assert "a" not in dg.keys()
    assert "b" in dg.keys()
    assert len(dg) == 1


def test_datagroup_equal_operator():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    assert Datagroup({"a": a, "b": b}) == Datagroup({"a": a, "b": b})


def test_datagroup_slice():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    expected = Datagroup(
        {"a": Array(values=[2.0], unit="m"), "b": Array(values=[7.0], unit="s")}
    )
    assert dg[1] == expected


def test_datagroup_slice_range():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    expected = Datagroup(
        {
            "a": Array(values=[2.0, 3.0, 4.0], unit="m"),
            "b": Array(values=[7.0, 8.0, 9.0], unit="s"),
        }
    )
    assert dg[1:4] == expected


def test_datagroup_sortby_key():
    a = Array(values=[2.0, 3.0, 1.0, 5.0, 4.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    dg.sortby("a")
    expected = Datagroup(
        {
            "a": Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m"),
            "b": Array(values=[8.0, 6.0, 7.0, 10.0, 9.0], unit="s"),
        }
    )
    assert arrayequal(dg["a"], expected["a"])
    assert arrayequal(dg["b"], expected["b"])


def test_datagroup_sortby_indices():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    dg.sortby([4, 3, 1, 2, 0])
    expected = Datagroup(
        {
            "a": Array(values=[5.0, 4.0, 2.0, 3.0, 1.0], unit="m"),
            "b": Array(values=[10.0, 9.0, 7.0, 8.0, 6.0], unit="s"),
        }
    )
    assert arrayequal(dg["a"], expected["a"])
    assert arrayequal(dg["b"], expected["b"])


def test_datagroup_clear():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    dg.clear()
    assert len(dg) == 0


def test_datagroup_get():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    c = Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="K")
    dg = Datagroup({"a": a, "b": b})
    assert arrayequal(dg.get("a", 42), a)
    assert dg.get("c", 42) == 42
    assert arrayequal(dg.get("c", c), c)


def test_datagroup_pop():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg = Datagroup({"a": a, "b": b})
    c = dg.pop("a")
    assert "a" not in dg
    assert arrayequal(c, a)


def test_datagroup_update():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    c = Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="K")
    dg = Datagroup({"a": a, "b": b})
    dg.update({"b": 2.0 * b, "c": c})
    assert arrayequal(dg["b"], Array(values=[12.0, 14.0, 16.0, 18.0, 20.0], unit="s"))
    assert arrayequal(dg["c"], c)


def test_copy():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = dg1.copy()
    del dg1["b"]
    assert "b" in dg2
    dg1["a"] *= 10.0
    assert arrayequal(dg2["a"], Array(values=[10.0, 20.0, 30.0, 40.0, 50.0], unit="m"))


def test_copy_overload():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = copy(dg1)
    del dg1["b"]
    assert "b" in dg2
    dg1["a"] *= 10.0
    assert arrayequal(dg2["a"], Array(values=[10.0, 20.0, 30.0, 40.0, 50.0], unit="m"))


def test_deepcopy():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = deepcopy(dg1)
    del dg1["b"]
    assert "b" in dg2
    dg1["a"] *= 10.0
    assert arrayequal(dg2["a"], Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m"))


def test_datagroup_shape():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    dg = Datagroup()
    dg["a"] = a
    assert dg.shape == (5,)
    a = Array(values=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], unit="m")
    dg = Datagroup()
    dg["a"] = a
    assert dg.shape == (2, 3)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
from common import arrayclose, arraytrue, arrayequal
from osyris import Array, units
from copy import copy, deepcopy
import numpy as np
from pint.errors import DimensionalityError
import pytest


def test_constructor_ndarray():
    a = np.arange(100.0)
    array = Array(values=a, unit="m")
    assert array.unit == units("m")
    assert len(array) == len(a)
    assert array.shape == a.shape
    assert np.array_equal(array.values, a)


def test_constructor_list():
    alist = [1.0, 2.0, 3.0, 4.0, 5.0]
    array = Array(values=alist, unit="s")
    assert array.unit == units("s")
    assert np.array_equal(array.values, alist)


def test_constructor_int():
    num = 15
    array = Array(values=num, unit="m")
    assert array.unit == units("m")
    assert np.array_equal(array.values, np.array(num))


def test_constructor_float():
    num = 154.77
    array = Array(values=num, unit="m")
    assert array.unit == units("m")
    assert np.array_equal(array.values, np.array(num))


def test_constructor_quantity():
    q = 6.7 * units("K")
    array = Array(values=q)
    assert array.unit == units("K")
    assert np.array_equal(array.values, np.array(q.magnitude))


def test_bad_constructor_quantity_with_unit():
    q = 6.7 * units("K")
    with pytest.raises(ValueError):
        _ = Array(values=q, unit="s")


def test_constructor_masked_array():
    a = np.arange(5.0)
    b = np.ma.masked_where(a > 2, a)
    array = Array(values=b, unit="m")
    assert array.unit == units("m")
    assert len(array) == len(b)
    assert array.shape == b.shape
    assert np.array_equal(array.values, b)
    assert np.array_equal(array.values.mask, [False, False, False, True, True])


def test_constructor_2d():
    a = np.arange(12.0).reshape(4, 3)
    array = Array(values=a, unit="m")
    assert array.unit == units("m")
    assert len(array) == len(a)
    assert array.shape == a.shape
    assert np.array_equal(array.values, a)


def test_addition():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[7.0, 9.0, 11.0, 13.0, 15.0], unit="m")
    assert arrayclose(a + b, expected)


def test_addition_conversion():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="cm")
    expected = Array(values=[1.06, 2.07, 3.08, 4.09, 5.1], unit="m")
    assert arrayclose(a + b, expected)


def test_addition_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    with pytest.raises(DimensionalityError):
        _ = a + b
    with pytest.raises(TypeError):
        _ = a + 3.0


def test_addition_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.5 * units("m")
    expected = Array(values=[4.5, 5.5, 6.5, 7.5, 8.5], unit="m")
    assert arrayclose(a + b, expected)


def test_addition_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[7.0, 9.0, 11.0, 13.0, 15.0], unit="m")
    a += b
    assert arrayclose(a, expected)


def test_addition_quantity_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.5 * units("m")
    expected = Array(values=[4.5, 5.5, 6.5, 7.5, 8.5], unit="m")
    a += b
    assert arrayclose(a, expected)


def test_subtraction():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[5.0, 5.0, 5.0, 5.0, 5.0], unit="m")
    assert arrayclose(b - a, expected)


def test_subtraction_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    with pytest.raises(DimensionalityError):
        _ = a - b
    with pytest.raises(TypeError):
        _ = a - 3.0


def test_subtraction_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.5 * units("m")
    expected = Array(values=[-2.5, -1.5, -0.5, 0.5, 1.5], unit="m")
    assert arrayclose(a - b, expected)


def test_subtraction_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[5.0, 5.0, 5.0, 5.0, 5.0], unit="m")
    b -= a
    assert arrayclose(b, expected)


def test_subtraction_quantity_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.5 * units("m")
    expected = Array(values=[-2.5, -1.5, -0.5, 0.5, 1.5], unit="m")
    a -= b
    assert arrayclose(a, expected)


def test_multiplication():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[6.0, 14.0, 24.0, 36.0, 50.0], unit="m*m")
    assert arrayclose(a * b, expected)


def test_multiplication_conversion():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="cm")
    expected = Array(values=[0.06, 0.14, 0.24, 0.36, 0.5], unit="m*m")
    assert arrayclose(a * b, expected)


def test_multiplication_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0
    expected = Array(values=[3.0, 6.0, 9.0, 12.0, 15.0], unit="m")
    assert arrayclose(a * b, expected)
    assert arrayclose(b * a, expected)


def test_multiplication_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = np.arange(5.0)
    expected = Array(values=[0.0, 2.0, 6.0, 12.0, 20.0], unit="m")
    assert arrayclose(a * b, expected)
    assert arrayclose(b * a, expected)


def test_multiplication_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.5 * units("s")
    expected = Array(values=[3.5, 7.0, 10.5, 14.0, 17.5], unit="m*s")
    assert arrayclose(a * b, expected)


def test_multiplication_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[6.0, 14.0, 24.0, 36.0, 50.0], unit="m*m")
    a *= b
    assert arrayclose(a, expected)


def test_multiplication_float_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0
    expected = Array(values=[3.0, 6.0, 9.0, 12.0, 15.0], unit="m")
    a *= b
    assert arrayclose(a, expected)


def test_multiplication_ndarray_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = np.arange(5.0)
    expected = Array(values=[0.0, 2.0, 6.0, 12.0, 20.0], unit="m")
    a *= b
    assert arrayclose(a, expected)


def test_multiplication_quantity_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.5 * units("s")
    expected = Array(values=[3.5, 7.0, 10.5, 14.0, 17.5], unit="m*s")
    a *= b
    assert arrayclose(a, expected)


def test_division():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[6.0, 3.5, 8.0 / 3.0, 2.25, 2.0], unit="m/s")
    assert arrayclose(b / a, expected)


def test_division_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = 3.0
    expected = Array(values=[1.0 / 3.0, 2.0 / 3.0, 1.0, 4.0 / 3.0, 5.0 / 3.0], unit="s")
    assert arrayclose(a / b, expected)
    expected = Array(values=[3.0, 3.0 / 2.0, 1.0, 3.0 / 4.0, 3.0 / 5.0], unit="1/s")
    assert arrayclose(b / a, expected)


def test_division_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = np.arange(5.0, 10.0)
    expected = Array(
        values=[1.0 / 5.0, 2.0 / 6.0, 3.0 / 7.0, 4.0 / 8.0, 5.0 / 9.0], unit="s"
    )
    assert arrayclose(a / b, expected)
    # expected = Array(values=[3., 3. / 2., 1., 3. / 4., 3. / 5.], unit='1/s')
    # assert arrayclose(b / a, expected)


def test_division_quantity():
    a = Array(values=[0.0, 2.0, 4.0, 6.0, 200.0], unit="s")
    b = 2.0 * units("s")
    expected = Array(values=[0.0, 1.0, 2.0, 3.0, 100.0], unit="dimensionless")
    assert arrayclose(a / b, expected)


def test_division_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="m")
    expected = Array(values=[6.0, 3.5, 8.0 / 3.0, 2.25, 2.0], unit="m/s")
    b /= a
    assert arrayclose(b, expected)


def test_division_float_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = 3.0
    expected = Array(values=[1.0 / 3.0, 2.0 / 3.0, 1.0, 4.0 / 3.0, 5.0 / 3.0], unit="s")
    a /= b
    assert arrayclose(a, expected)


def test_division_ndarray_inplace():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = np.arange(5.0, 10.0)
    expected = Array(
        values=[1.0 / 5.0, 2.0 / 6.0, 3.0 / 7.0, 4.0 / 8.0, 5.0 / 9.0], unit="s"
    )
    a /= b
    assert arrayclose(a, expected)
    # expected = Array(values=[3., 3. / 2., 1., 3. / 4., 3. / 5.], unit='1/s')
    # assert arrayclose(b / a, expected)


def test_division_quantity_inplace():
    a = Array(values=[0.0, 2.0, 4.0, 6.0, 200.0], unit="s")
    b = 2.0 * units("s")
    expected = Array(values=[0.0, 1.0, 2.0, 3.0, 100.0], unit="dimensionless")
    a /= b
    assert arrayclose(a, expected)


def test_power():
    a = Array(values=[1.0, 2.0, 4.0, 6.0, 200.0], unit="s")
    expected = Array(values=[1.0, 8.0, 64.0, 216.0, 8.0e6], unit="s**3")
    assert arrayclose(a**3, expected)


def test_negative():
    a = Array(values=[1.0, 2.0, 4.0, 6.0, 200.0], unit="s")
    expected = Array(values=[-1.0, -2.0, -4.0, -6.0, -200.0], unit="s")
    assert arrayequal(-a, expected)


def test_equal():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[11.0, 2.0, 3.0, 4.1, 5.0], unit="m")
    expected = [False, True, True, False, True]
    assert all((a == b).values == expected)


def test_equal_conversion():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[1100.0, 200.0, 300.0, 410.0, 500.0], unit="cm")
    expected = [False, True, True, False, True]
    assert all((a == b).values == expected)


def test_equal_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[11.0, 2.0, 3.0, 4.1, 5.0], unit="s")
    with pytest.raises(DimensionalityError):
        _ = a == b


def test_equal_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = np.array([11.0, 2.0, 3.0, 4.1, 5.0])
    expected = [False, True, True, False, True]
    assert all((a == b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a == b


def test_equal_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = 3.0
    expected = [False, False, True, False, False]
    assert all((a == b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a == b


def test_equal_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0 * units("m")
    expected = [False, False, True, False, False]
    assert all((a == b).values == expected)
    b = 3.0 * units("s")
    with pytest.raises(DimensionalityError):
        _ = a == b


def test_not_equal():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[11.0, 2.0, 3.0, 4.1, 5.0], unit="m")
    expected = [True, False, False, True, False]
    assert all((a != b).values == expected)


def test_not_equal_conversion():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[1100.0, 200.0, 300.0, 410.0, 500.0], unit="cm")
    expected = [True, False, False, True, False]
    assert all((a != b).values == expected)


def test_not_equal_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[11.0, 2.0, 3.0, 4.1, 5.0], unit="s")
    with pytest.raises(DimensionalityError):
        _ = a != b


def test_not_equal_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = np.array([11.0, 2.0, 3.0, 4.1, 5.0])
    expected = [True, False, False, True, False]
    assert all((a != b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a != b


def test_not_equal_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = 3.0
    expected = [True, True, False, True, True]
    assert all((a != b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a != b


def test_not_equal_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0 * units("m")
    expected = [True, True, False, True, True]
    assert all((a != b).values == expected)
    b = 3.0 * units("s")
    with pytest.raises(DimensionalityError):
        _ = a != b


def test_less_than():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="s")
    expected = [True, True, False, False, True]
    assert all((a < b).values == expected)


def test_less_than_conversion():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[600.0, 700.0, 100.0, 400.0, 1000.0], unit="cm")
    expected = [True, True, False, False, True]
    assert all((a < b).values == expected)


def test_less_than_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="m")
    with pytest.raises(DimensionalityError):
        _ = a < b


def test_less_than_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = np.array([6.0, 7.0, 1.0, 4.0, 10.0])
    expected = [True, True, False, False, True]
    assert all((a < b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a < b


def test_less_than_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = 3.0
    expected = [True, True, False, False, False]
    assert all((a < b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a < b


def test_less_than_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0 * units("m")
    expected = [True, True, False, False, False]
    assert all((a < b).values == expected)
    b = 3.0 * units("s")
    with pytest.raises(DimensionalityError):
        _ = a < b


def test_less_equal():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="s")
    expected = [True, True, False, True, True]
    assert all((a <= b).values == expected)


def test_less_equal_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="m")
    with pytest.raises(DimensionalityError):
        _ = a <= b


def test_less_equal_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = np.array([6.0, 7.0, 1.0, 4.0, 10.0])
    expected = [True, True, False, True, True]
    assert all((a <= b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a < b


def test_less_equal_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = 3.0
    expected = [True, True, True, False, False]
    assert all((a <= b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a < b


def test_less_equal_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0 * units("m")
    expected = [True, True, True, False, False]
    assert all((a <= b).values == expected)
    b = 3.0 * units("s")
    with pytest.raises(DimensionalityError):
        _ = a < b


def test_greater_than():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="s")
    expected = [True, True, False, False, True]
    assert all((b > a).values == expected)


def test_greater_than_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="K")
    with pytest.raises(DimensionalityError):
        _ = b > a


def test_greater_than_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = np.array([6.0, 7.0, 1.0, 4.0, 10.0])
    expected = [False, False, True, False, False]
    assert all((a > b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a > b


def test_greater_than_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = 3.0
    expected = [False, False, False, True, True]
    assert all((a > b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a > b


def test_greater_than_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0 * units("m")
    expected = [False, False, False, True, True]
    assert all((a > b).values == expected)
    b = 3.0 * units("s")
    with pytest.raises(DimensionalityError):
        _ = a > b


def test_greater_equal():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="s")
    expected = [True, True, False, True, True]
    assert all((b >= a).values == expected)


def test_greater_equal_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="s")
    b = Array(values=[6.0, 7.0, 1.0, 4.0, 10.0], unit="K")
    with pytest.raises(DimensionalityError):
        _ = b >= a


def test_greater_equal_ndarray():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = np.array([6.0, 7.0, 1.0, 4.0, 10.0])
    expected = [False, False, True, True, False]
    assert all((a >= b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a >= b


def test_greater_equal_float():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0])
    b = 3.0
    expected = [False, False, True, True, True]
    assert all((a >= b).values == expected)
    a.unit = "m"
    with pytest.raises(DimensionalityError):
        _ = a >= b


def test_greater_equal_quantity():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = 3.0 * units("m")
    expected = [False, False, True, True, True]
    assert all((a >= b).values == expected)
    b = 3.0 * units("s")
    with pytest.raises(DimensionalityError):
        _ = a >= b


def test_logical_and():
    a = Array(values=[True, True, True, False, False, False])
    b = Array(values=[True, False, True, False, True, False])
    expected = [True, False, True, False, False, False]
    assert all((b & a).values == expected)


def test_logical_or():
    a = Array(values=[True, True, True, False, False, False])
    b = Array(values=[True, False, True, False, True, False])
    expected = [True, True, True, False, True, False]
    assert all((b | a).values == expected)


def test_logical_xor():
    a = Array(values=[True, True, True, False, False, False])
    b = Array(values=[True, False, True, False, True, False])
    expected = [False, True, False, False, True, False]
    assert all((b ^ a).values == expected)


def test_logical_invert():
    a = Array(values=[True, True, False, False, True, False])
    expected = [False, False, True, True, False, True]
    assert all((~a).values == expected)


def test_to():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[1.0e-3, 2.0e-3, 3.0e-3, 4.0e-3, 5.0e-3], unit="km")
    assert arrayclose(a.to("km"), b)
    assert a.unit == units("m")


def test_to_bad_units():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    with pytest.raises(DimensionalityError):
        _ = a.to("s")


def test_min():
    a = Array(values=[1.0, -2.0, 3.0, 0.4, 0.5, 0.6], unit="m")
    assert (a.min() == Array(values=-2.0, unit="m")).values
    b = Array(values=np.array([1.0, -2.0, 3.0, 0.4, 0.5, 0.6]).reshape(2, 3), unit="m")
    assert (b.min() == Array(values=-2.0, unit="m")).values


def test_max():
    a = Array(values=[1.0, 2.0, 3.0, -15.0, 5.0, 6.0], unit="m")
    assert (a.max() == Array(values=6.0, unit="m")).values
    b = Array(values=np.array([1.0, 2.0, 3.0, -15.0, 5.0, 6.0]).reshape(2, 3), unit="m")
    assert (b.max() == Array(values=6.0, unit="m")).values


def test_reshape():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0], unit="m")
    expected = Array(values=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], unit="m")
    assert arrayequal(a.reshape(2, 3), expected)


def test_slicing():
    a = Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="m")
    assert a[2] == Array(values=[13.0], unit="m")
    assert arrayequal(a[:4], Array(values=[11.0, 12.0, 13.0, 14.0], unit="m"))
    assert arrayequal(a[2:4], Array(values=[13.0, 14.0], unit="m"))


def test_slicing_2d():
    a = Array(values=np.arange(12.0).reshape(4, 3), unit="m")
    assert arrayequal(a[0, :], Array(values=[0.0, 1.0, 2.0], unit="m"))
    assert arrayequal(a[1], Array(values=[3.0, 4.0, 5.0], unit="m"))
    assert arrayequal(a[2:3], Array(values=[[6.0, 7.0, 8.0]], unit="m"))
    assert a[2:3].shape == (1, 3)
    assert arrayequal(a[:2], Array(values=[[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]], unit="m"))


def test_slicing_with_array():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    select = a > (3.0 * units("m"))
    assert arrayequal(a[select], Array(values=[4.0, 5.0], unit="m"))
    b = Array(values=[0, 2, 4])
    assert arrayequal(a[b], Array(values=[1.0, 3.0, 5.0], unit="m"))


def test_slicing_with_array_bad_dtype():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    with pytest.raises(TypeError):
        _ = a[Array(values=[0.0, 2.0, 4.0])]


def test_copy():
    a = Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="m")
    b = a.copy()
    a *= 10.0
    assert arraytrue(b == Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="m"))


def test_copy_overload():
    a = Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="m")
    b = copy(a)
    a *= 10.0
    assert arraytrue(b == Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="m"))


def test_deepcopy():
    a = Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="m")
    b = deepcopy(a)
    a *= 10.0
    assert arraytrue(b == Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="m"))


def test_numpy_unary():
    values = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=values, unit="m")
    expected = np.log10(values)
    result = np.log10(a)
    assert np.allclose(result.values, expected)
    assert result.unit == units("m")


def test_numpy_sqrt():
    values = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=values, unit="m*m")
    expected = np.sqrt(values)
    result = np.sqrt(a)
    assert np.allclose(result.values, expected)
    assert result.unit == units("m")


def test_numpy_binary():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    b_buf = [6.0, 7.0, 8.0, 9.0, 10.0]
    a = Array(values=a_buf, unit="m")
    b = Array(values=b_buf, unit="m")
    expected = np.dot(a_buf, b_buf)
    result = np.dot(a, b)
    assert result.values == expected
    assert result.unit == units("m")


def test_numpy_iterable():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    b_buf = [6.0, 7.0, 8.0, 9.0, 10.0]
    a = Array(values=a_buf, unit="m")
    b = Array(values=b_buf, unit="m")
    expected = np.concatenate([a_buf, b_buf])
    result = np.concatenate([a, b])
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m")


def test_numpy_multiply_with_ndarray():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=a_buf, unit="m")
    b = np.array([6.0, 7.0, 8.0, 9.0, 10.0])
    expected = np.multiply(a_buf, b)
    result = np.multiply(a, b)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m")
    result = np.multiply(b, a)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m")


def test_numpy_multiply_with_quantity():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=a_buf, unit="m")
    b = 3.5 * units("s")
    expected = np.multiply(a_buf, b.magnitude)
    result = np.multiply(a, b)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m*s")


def test_numpy_multiply_with_float():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=a_buf, unit="m")
    b = 3.5
    expected = np.multiply(a_buf, b)
    result = np.multiply(a, b)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m")
    result = np.multiply(b, a)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m")


def test_numpy_divide_with_ndarray():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=a_buf, unit="m")
    b = np.array([6.0, 7.0, 8.0, 9.0, 10.0])
    expected = np.divide(a_buf, b)
    result = np.divide(a, b)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m")
    expected = np.divide(b, a_buf)
    result = np.divide(b, a)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("1/m")


def test_numpy_divide_with_quantity():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=a_buf, unit="m")
    b = 3.5 * units("s")
    expected = np.divide(a_buf, b.magnitude)
    result = np.divide(a, b)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m/s")


def test_numpy_divide_with_float():
    a_buf = [1.0, 2.0, 3.0, 4.0, 5.0]
    a = Array(values=a_buf, unit="m")
    b = 3.5
    expected = np.divide(a_buf, b)
    result = np.divide(a, b)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("m")
    expected = np.divide(b, a_buf)
    result = np.divide(b, a)
    assert np.array_equal(result.values, expected)
    assert result.unit == units("1/m")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Osyris contributors (https://github.com/osyris-project/osyris)
from common import arrayequal
from osyris import Array, Datagroup, Dataset
from copy import copy, deepcopy


def test_dataset_creation():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    ds = Dataset()
    ds["one"] = dg1
    ds["two"] = dg2
    assert "one" in ds
    assert "two" in ds


def test_dataset_clear():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    ds = Dataset()
    ds["one"] = dg1
    ds["two"] = dg2
    ds.clear()
    assert len(ds) == 0


def test_dataset_get():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    ds = Dataset()
    ds["one"] = dg1
    ds["two"] = dg2
    assert ds.get("one", 42) == dg1
    assert ds.get("three", 42) == 42


def test_dataset_pop():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    ds = Dataset()
    ds["one"] = dg1
    ds["two"] = dg2
    c = ds.pop("one")
    assert "one" not in ds
    assert c == dg1


def test_dataset_update():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    c = Array(values=[11.0, 12.0, 13.0, 14.0, 15.0], unit="K")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    dg3 = Datagroup({"a": a, "c": c})
    ds = Dataset()
    ds["one"] = dg1
    ds["two"] = dg2
    ds.update({"one": Datagroup({"a": a * 2.0}), "three": dg3})
    assert "three" in ds
    assert "b" not in ds["one"]
    assert "c" in ds["three"]


def test_copy():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    ds1 = Dataset()
    ds1["one"] = dg1
    ds1["two"] = dg2
    ds1.meta = {"metadata1": [1, 2, 3], "metadata2": "4,5,6"}
    ds2 = ds1.copy()
    del ds1["two"]
    assert "two" in ds2
    del dg1["b"]
    assert "b" not in ds2["one"]
    dg1["a"] *= 10.0
    assert arrayequal(
        ds2["one"]["a"], Array(values=[10.0, 20.0, 30.0, 40.0, 50.0], unit="m")
    )
    del ds1.meta["metadata2"]
    assert "metadata2" in ds2.meta
    ds1.meta["metadata1"][0] = 0
    assert ds2.meta["metadata1"][0] == 0


def test_copy_overload():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    ds1 = Dataset()
    ds1["one"] = dg1
    ds1["two"] = dg2
    ds1.meta = {"metadata1": [1, 2, 3], "metadata2": "4,5,6"}
    ds2 = copy(ds1)
    del ds1["two"]
    assert "two" in ds2
    del dg1["b"]
    assert "b" not in ds2["one"]
    dg1["a"] *= 10.0
    assert arrayequal(
        ds2["one"]["a"], Array(values=[10.0, 20.0, 30.0, 40.0, 50.0], unit="m")
    )
    del ds1.meta["metadata2"]
    assert "metadata2" in ds2.meta
    ds1.meta["metadata1"][0] = 0
    assert ds2.meta["metadata1"][0] == 0


def test_deepcopy():
    a = Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    b = Array(values=[6.0, 7.0, 8.0, 9.0, 10.0], unit="s")
    dg1 = Datagroup({"a": a, "b": b})
    dg2 = Datagroup({"a": a[:-1], "b": b[:-1]})
    ds1 = Dataset()
    ds1["one"] = dg1
    ds1["two"] = dg2
    ds1.meta = {"metadata1": [1, 2, 3], "metadata2": "4,5,6"}
    ds2 = deepcopy(ds1)
    del ds1["two"]
    assert "two" in ds2
    del dg1["b"]
    assert "b" in ds2["one"]
    dg1["a"] *= 10.0
    assert arrayequal(
        ds2["one"]["a"], Array(values=[1.0, 2.0, 3.0, 4.0, 5.0], unit="m")
    )
    del ds1.meta["metadata2"]
    assert "metadata2" in ds2.meta
    ds1.meta["metadata1"][0] = 0
    assert ds2.meta["metadata1"][0] == 1
